NAME = "polyaxon"
VERSION = "2.0.0-rc54"
SCHEMA_VERSION = 1.1
DESC = "Command Line Interface (CLI) and client to interact with Polyaxon API."
URL = "https://github.com/polyaxon/polyaxon"
AUTHOR = "Polyaxon, Inc."
EMAIL = "contact@polyaxon.com"
LICENSE = "Apache 2.0"
